#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Default initialization for flow tests
# Adrien Crovato

from fwk.wutils import parseargs
import flow as flo
import tbox

def mesh(dim, file, pars, bnd, wk = 'wake', wktp = None):
    '''Initialize mesh and mesh writer
    '''
    import tbox.gmsh as gmsh
    # load the mesh
    msh = gmsh.MeshLoader(file,__file__).execute(**pars)
    # crack the mesh
    mshCrck = tbox.MshCrack(msh, dim)
    mshCrck.setCrack(wk)
    mshCrck.addBoundaries(bnd)
    if dim == 3 and wktp is not None:
        mshCrck.setExcluded(wktp)
    mshCrck.run()
    del mshCrck
    # create writer
    gmshWriter = tbox.GmshExport(msh)
    gmshWriter.save(msh.name)
    return msh, gmshWriter

def problem(msh, dim, alpha, beta, minf, sref, lref, xref, yref, zref, sur, fld = 'field', upstr = 'upstream', ff = 'farfield', dnstr = 'downstream', wk = 'wake', te = None, tp = None, vsc = False, dbc = False):
    '''Initialize problem
    '''
    # create base formulation
    pbl = flo.Problem(msh, dim, alpha, beta, minf, sref, lref, xref, yref, zref)
    # add incompressible or compressible air medium
    pbl.set(flo.Medium(msh, fld, minf, dim, alpha, beta))
    # add initial condition
    pbl.set(flo.Initial(msh, fld, dim, alpha, beta))
    if dbc:
        # Dirichlet BCs provided and explicitely imposed
        dirichlet = flo.Dirichlet(msh, upstr, dim, alpha, beta)
        pbl.add(dirichlet)
    else:
        # Dirichlet BCs not provided, Neumann imposed instead, and a DOF will be pinned automatically
        dirichlet = None
        pbl.add(flo.Freestream(msh, upstr, dim, alpha, beta))
    # add freestream (Neumann) conditions
    pbl.add(flo.Freestream(msh, ff, dim, alpha, beta))
    pbl.add(flo.Freestream(msh, dnstr, dim, alpha, beta))
    # add wake and Kutta conditions
    if dim == 2 and te is not None:
        wake = flo.Wake(msh, [wk, wk+'_', fld])
        pbl.add(wake)
        pbl.add(flo.Kutta(msh, [te, wk+'_', sur, fld]))
    elif dim == 3:
        wake = flo.Wake(msh, [wk, wk+'_', fld, tp])
        pbl.add(wake)
    else:
        wake = None
    # add boundary of interest
    bnd = flo.Boundary(msh, [sur, fld])
    pbl.add(bnd)
    # add transpiration BC (VII)
    if vsc:
        blw = flo.Blowing(msh, sur)
        blww = flo.Blowing(msh, wk)
        pbl.add(blw)
        pbl.add(blww)
    else:
        blw = None
        blww = None
    return pbl, dirichlet, wake, bnd, [blw, blww]

def picard(pbl):
    '''Initialize Picard solver
    '''
    args = parseargs()
    solver = flo.Picard(pbl, tbox.Gmres())
    solver.nthreads = args.k
    solver.verbose = args.verb
    solver.relTol = 1e-6
    solver.absTol = 1e-8
    solver.maxIt = 25
    solver.relax = 0.7
    return solver

def newton(pbl):
    '''Initialize Newton solver
    '''
    from tbox.solvers import LinearSolver
    args = parseargs()
    solver = flo.Newton(pbl, LinearSolver().pardiso())
    solver.nthreads = args.k
    solver.verbose = args.verb
    solver.relTol = 1e-6
    solver.absTol = 1e-8
    solver.maxIt = 25
    solver.lsTol = 1e-6
    solver.maxLsIt = 10
    solver.avThrsh = 1e-2
    return solver

def meshMorpher(msh, dim, mov, fxd = ['upstream', 'farfield', 'downstream'], fld = 'field', wk = 'wake', sym = 'symmetry'):
    '''Initialize mesh morpher
    '''
    args = parseargs()
    mshDef = tbox.MshDeform(msh, dim, args.k)
    mshDef.setField(fld)
    mshDef.addFixed(fxd)
    mshDef.addMoving(mov)
    if dim == 3:
        mshDef.setSymmetry(sym, 1)
    mshDef.addInternal([wk, wk+'_'])
    mshDef.initialize()
    return mshDef

def adjoint(solver, morpher):
    '''Initialize adjoint solver
    '''
    from tbox.solvers import LinearSolver
    adjoint = flo.Adjoint(solver, morpher, LinearSolver().pardiso())
    return adjoint

def initViewer(problem):
    '''Initialize viewer
    '''
    args = parseargs()
    if not args.nogui:
        from tbox.viewer import GUI
        GUI().open( problem.msh.name )
