/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wAssign.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>
using namespace tbox;
using namespace flow;

Assign::Assign(std::shared_ptr<MshData> _msh, int no) : Group(_msh, no)
{
    getNodes();
}
Assign::Assign(std::shared_ptr<MshData> _msh, std::string const &name) : Group(_msh, name)
{
    getNodes();
}
Assign::~Assign()
{
    delete f;
    std::cout << "~Assign()\n";
}

/**
 * @brief Prescribe value to all nodes (default implementation)
 */
void Assign::apply(std::vector<double> &vec)
{
    for (auto n : nodes)
        vec[n->row] = f->eval(n->pos);
}

/**
 * @brief Get all nodes belonging to boundary tag
 */
void Assign::getNodes()
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);
    std::sort(nodes.begin(), nodes.end());
    auto it = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), it));
}

Initial::Initial(std::shared_ptr<tbox::MshData> _msh, int no, int dim, double alpha, double beta) : Assign(_msh, no)
{
    f = new F0PsPhiInf(dim, alpha, beta);
}
Initial::Initial(std::shared_ptr<tbox::MshData> _msh, std::string const &name, int dim, double alpha, double beta) : Assign(_msh, name)
{
    f = new F0PsPhiInf(dim, alpha, beta);
}
void Initial::write(std::ostream &out) const
{
    out << "flow::Initial condition on " << *tag << ")\n";
    for (auto n : nodes)
        out << '\t' << *n << ": (val=" << f->eval(n->pos) << ")\n";
}

Dirichlet::Dirichlet(std::shared_ptr<tbox::MshData> _msh, int no, int dim, double alpha, double beta, bool pin) : Assign(_msh, no)
{
    f = new F0PsPhiInf(dim, alpha, beta);
    // Only retain the first node, so that the DOF associated to this node will be pinned
    if (pin)
        this->nodes.resize(1);
}
Dirichlet::Dirichlet(std::shared_ptr<tbox::MshData> _msh, std::string const &name, int dim, double alpha, double beta) : Assign(_msh, name)
{
    f = new F0PsPhiInf(dim, alpha, beta);
}
void Dirichlet::write(std::ostream &out) const
{
    out << "flow::Dirichlet boundary condition on " << *tag << ")\n";
    for (auto n : nodes)
        out << '\t' << *n << ": (val=" << f->eval(n->pos) << ")\n";
}