/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wWakeResidual.h"
#include "wWakeElement.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"

using namespace tbox;
using namespace flow;

/**
 * @brief Build the residual matrix for a fixed-point iteration, on one wake element
 *
 * A_u = \int (psi + grad_psi*v_inf) * v_u*grad_phi_u dS
 * A_l = \int (psi + grad_psi*v_inf) * v_l*grad_phi_l dS
 * NB: A_u and A_l will be assembled on same row DOF, but on different columns
 * NB: v_inf = [1, 0, 0]
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> WakeResidual::buildFixed(WakeElement const &we, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &sCache = we.surUpE->getVCache();
    Gauss &sGauss = sCache.getVGauss();
    Eigen::MatrixXd const &vUpDsf = we.volUpE->getVCache().getDsf(0);
    Eigen::MatrixXd const &vLwDsf = we.volLwE->getVCache().getDsf(0);
    // Get Jacobian
    Eigen::MatrixXd const &vUpJ = we.volUpE->getJinv(0);
    Eigen::MatrixXd const &vLwJ = we.volLwE->getJinv(0);
    // Get size
    size_t nRow = we.nRow;
    size_t nDim = we.nDim;

    // Stabilization
    Eigen::MatrixXd dSf(nRow, nDim);
    for (size_t i = 0; i < nRow; ++i)
        for (size_t j = 0; j < nDim; ++j)
            dSf(i, j) = vUpDsf(j, we.vsMap.at(we.wkN[i].first));
    Eigen::VectorXd dSfp(nRow);
    dSfp = 0.5 * sqrt(we.surUpE->getVol()) * dSf * vUpJ.transpose() * Eigen::Vector3d(1, 0, 0).block(0, 0, nDim, 1);
    // Velocity
    Eigen::RowVectorXd Aup = we.volUpE->computeGradient(phi, 0).transpose() * vUpJ * vUpDsf;
    Eigen::RowVectorXd Alw = we.volLwE->computeGradient(phi, 0).transpose() * vLwJ * vLwDsf;

    // Build
    Eigen::MatrixXd Au = Eigen::MatrixXd::Zero(nRow, we.nColUp);
    Eigen::MatrixXd Al = Eigen::MatrixXd::Zero(nRow, we.nColLw);
    for (size_t k = 0; k < sGauss.getN(); ++k)
    {
        // stabilized shape functions
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = dSfp(i) + sf(we.wkN[i].first);
        // wake contribution
        Au += sfp * Aup * sGauss.getW(k) * we.surUpE->getDetJ(k);
        Al += sfp * Alw * sGauss.getW(k) * we.surUpE->getDetJ(k);
    }
    return std::make_tuple(Au, Al);
}

/**
 * @brief Build the residual vector, on one wake element
 *
 * b_u = \int (psi + grad_psi*v_inf) * (grad_phi_u)^2 dS
 * b_l = \int (psi + grad_psi*v_inf) * (grad_phi_l)^2 dS
 * NB: v_inf = [1, 0, 0]
 */
std::tuple<Eigen::VectorXd, Eigen::VectorXd> WakeResidual::build(WakeElement const &we, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &sCache = we.surUpE->getVCache();
    Gauss &sGauss = sCache.getVGauss();
    Eigen::MatrixXd const &vUpDsf = we.volUpE->getVCache().getDsf(0);
    // Get size
    size_t nRow = we.nRow;
    size_t nDim = we.nDim;

    // Stabilization
    Eigen::MatrixXd dSf(nRow, nDim);
    for (size_t i = 0; i < nRow; ++i)
        for (size_t j = 0; j < nDim; ++j)
            dSf(i, j) = vUpDsf(j, we.vsMap.at(we.wkN[i].first));
    Eigen::VectorXd dSfp(nRow);
    dSfp = 0.5 * sqrt(we.surUpE->getVol()) * dSf * we.volUpE->getJinv(0).transpose() * Eigen::Vector3d(1, 0, 0).block(0, 0, nDim, 1);
    // Velocity squared
    Eigen::VectorXd dPhiU = we.volUpE->computeGradient(phi, 0);
    Eigen::VectorXd dPhiL = we.volLwE->computeGradient(phi, 0);
    double dPhi2u = dPhiU.squaredNorm();
    double dPhi2l = dPhiL.squaredNorm();

    // Build
    Eigen::VectorXd bu = Eigen::VectorXd::Zero(nRow);
    Eigen::VectorXd bl = Eigen::VectorXd::Zero(nRow);
    for (size_t k = 0; k < sGauss.getN(); ++k)
    {
        // stabilized shape functions
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = dSfp(i) + sf(we.wkN[i].first);
        // wake contribution
        bu += sfp * dPhi2u * sGauss.getW(k) * we.surUpE->getDetJ(k);
        bl += sfp * dPhi2l * sGauss.getW(k) * we.surUpE->getDetJ(k);
    }
    return std::make_tuple(bu, bl);
}

/**
 * @brief Build the gradient of the residual vector with respect to the flow variable (jacobian), on one wake element
 *
 * A_u = \int (psi + grad_psi*v_inf) * 2*(grad_phi_u)*dgrad_phi_u dS
 * A_l = \int (psi + grad_psi*v_inf) * 2*(grad_phi_l)*dgrad_phi_l dS
 * NB: v_inf = [1, 0, 0]
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> WakeResidual::buildGradientFlow(WakeElement const &we, std::vector<double> const &phi)
{
    // Build
    Eigen::MatrixXd Au, Al;
    std::tie(Au, Al) = WakeResidual::buildFixed(we, phi);
    return std::make_tuple(2 * Au, 2 * Al);
}

/**
 * @brief Build the gradient of the residual vector with respect to the nodes, on one wake element
 *
 * A = d( \int (psi + grad_psi*v_inf) * (grad_phi)^2 dS )
 *   = \int dgrad_psi*v_inf * (grad_phi)^2 dS
 *   + \int (psi + grad_psi*v_inf) * d(grad_phi)^2 dS
 *   + \int (psi + grad_psi*v_inf) * (grad_phi)^2 ddS
 * NB: v_inf = [1, 0, 0]
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd, Eigen::MatrixXd> WakeResidual::buildGradientMesh(WakeElement const &we, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &sCache = we.surUpE->getVCache();
    Gauss &surGauss = sCache.getVGauss();
    Eigen::MatrixXd const &dSfUp = we.volUpE->getVCache().getDsf(0);
    // Get Jacobian
    Eigen::MatrixXd const &iJUp = we.volUpE->getJinv(0);
    Eigen::MatrixXd const &iJLw = we.volLwE->getJinv(0);
    std::vector<Eigen::MatrixXd> const &dJUp = we.volUpE->getGradJ(0);
    std::vector<Eigen::MatrixXd> const &dJLw = we.volLwE->getGradJ(0);
    // Get surface gradient
    std::vector<double> const &dSurf = we.surUpE->getGradVol();
    // Get size
    size_t nRow = we.nRow;
    size_t nDim = we.nDim;
    size_t nColUp = we.nColUp;
    size_t nColLw = we.nColLw;
    size_t nSur = we.surUpE->nodes.size();

    // Stabilization term
    Eigen::VectorXd vi = Eigen::Vector3d(1, 0, 0).block(0, 0, nDim, 1);
    double sqSurf = sqrt(we.surUpE->getVol());
    Eigen::MatrixXd dSf(nRow, nDim);
    for (size_t i = 0; i < nRow; ++i)
        for (size_t m = 0; m < nDim; ++m)
            dSf(i, m) = dSfUp(m, we.vsMap.at(we.wkN[i].first));
    Eigen::VectorXd gradSfp = 0.5 * sqSurf * dSf * iJUp.transpose() * vi;
    // Gradient of stabilization term (volume and surface)
    std::vector<Eigen::VectorXd> dGradSfV(nDim * nColUp);
    for (size_t i = 0; i < nColUp; ++i)
    {
        for (size_t m = 0; m < nDim; ++m)
        {
            size_t idx = i * nDim + m;
            dGradSfV[idx] = 0.5 * sqSurf * dSf * (-iJUp * dJUp[idx] * iJUp).transpose() * vi; // S * dgrad_psi
        }
    }
    std::vector<Eigen::VectorXd> dGradSfS(nDim * nSur);
    for (size_t i = 0; i < nSur; ++i)
    {
        for (size_t m = 0; m < nDim; ++m)
        {
            size_t idx = i * nDim + m;
            dGradSfS[idx] = 0.5 * 0.5 / sqSurf * dSurf[idx] * dSf * iJUp.transpose() * vi; // dS * grad_psi
        }
    }
    // Gradients of potential
    Eigen::VectorXd dPhiUp = we.volUpE->computeGradient(phi, 0);
    Eigen::VectorXd dPhiLw = we.volLwE->computeGradient(phi, 0);

    // Build
    Eigen::MatrixXd Au = Eigen::MatrixXd::Zero(nRow, nDim * nColUp);
    Eigen::MatrixXd Al = Eigen::MatrixXd::Zero(nRow, nDim * nColLw);
    Eigen::MatrixXd As = Eigen::MatrixXd::Zero(nRow, nDim * nSur);
    for (size_t k = 0; k < surGauss.getN(); ++k)
    {
        // weight and Jacobian determinant
        double w = surGauss.getW(k);
        double detJ = we.surUpE->getDetJ(k);
        std::vector<double> dDetJ = we.surUpE->getGradDetJ(k);
        // shape functions and stabilization
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(we.wkN[i].first) + gradSfp(i);

        // Build gradient wrt to upper volume nodes
        for (size_t i = 0; i < nColUp; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of velocity jump
                Au.col(idx) += w * sfp * 2 * dPhiUp.transpose() * (-iJUp * dJUp[idx] * dPhiUp) * detJ; // (psi + grad_psi) * dgrad_phi^2 * detJ
                // gradient of stabilization term (volume)
                Au.col(idx) += w * dGradSfV[idx] * (dPhiUp.dot(dPhiUp) - dPhiLw.dot(dPhiLw)) * detJ; // dgrad_psi' * grad_phi^2 * detJ
            }
        }
        // Build gradient wrt to lower volume nodes
        for (size_t i = 0; i < nColLw; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of velocity jump
                Al.col(idx) += w * sfp * 2 * dPhiLw.transpose() * (-iJLw * dJLw[idx] * dPhiLw) * detJ; // (psi + grad_psi) * dgrad_phi^2 * detJ
            }
        }
        // Build gradient wrt to (upper) surface nodes
        for (size_t i = 0; i < nSur; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of stabilization term (surface)
                As.col(idx) += w * dGradSfS[idx] * (dPhiUp.dot(dPhiUp) - dPhiLw.dot(dPhiLw)) * detJ; // dgrad_psi * grad_phi^2 * detJ
                // gradient of jacobian determinant
                As.col(idx) += w * sfp * (dPhiUp.dot(dPhiUp) - dPhiLw.dot(dPhiLw)) * dDetJ[idx]; // (N + grad_psi) * grad_phi^2 * ddetJ
            }
        }
    }
    return std::make_tuple(Au, Al, As);
}
